from opengridview.parser import Parser
import re

class ParserAutoSplit(Parser):

  def __init__(self,config):
    Parser.__init__(self,config) # Base
    if config.has_key('separator'):
      self.separator = config['separator']
    else:
      self.separator = '\s+'

  def read_stream(self, stream, item_func, header_func):
    while True:
      line = stream.readline()
      if line:
        line = line.rstrip()
        row = re.split(self.separator,line)
        if not self.header_names:
          # This is the first row. Use it to set header names
          self.set_headers(row)
          continue
        else:
          self.add_item(row,header_func,item_func)
      else:
        break



from gi.repository import Gtk
from opengridview.window import Window
from gi.repository import GObject
import sys

windows = []

def window_closed(window):
    global windows
    windows.remove(window)
    if not windows:
        Gtk.main_quit()

def main():
    """Main entry point"""

    # Quit when interrupted
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Arguments:
    # --separator STRING/REGEX - how to split a row into cells (only relevant for CSV parser)
    # --flatten - flatten item hashes. {'a':{'b':'c'}} --> {'a_b':'c'}
    import argparse
    parser = argparse.ArgumentParser(description='View tabulated data via GUI')
    parser.add_argument('-p','--parser',type=str, default='autosplit',help='Type of parser to use') #TODO add possible parsers
    parser.add_argument('--headers',type=str, help='Headers are this comma-delimited names instead of ones supplied in file. Use colons to specify types, like "colname:int"')
    parser.add_argument('--filter',type=str, help='Pre-populate filter box')
    parser.add_argument('-s', '--separator', help='How to seperate columns. Applies only to some parsers')
    parser.add_argument('files', nargs='*', help='Files to show. Each file opens a new window')
    args = parser.parse_args()

    GObject.threads_init()

    default_config = {'format': args.parser}
    if args.filter:
        default_config['filter']=args.filter
    if args.separator:
        default_config['separator']=args.separator
    if args.headers:
        default_config['headers']=map(lambda s: s.strip(),args.headers.split(','))
    inputs = [ (open(f,'r'),default_config) for f in args.files ]
    # Add stdin as input, if it's not a tty
    if not sys.stdin.isatty():
        inputs.append((sys.stdin, default_config))

    global windows
    windows = [Window(i[0],i[1]) for i in inputs]
    for win in windows:
      win.start_read()
      win.connect("destroy",window_closed)
    if windows:
        Gtk.main()
    else:
        print 'No input supplied so no windows are created'




class Parser:

  @staticmethod
  def find_parser(config):
    parser = config['format']
    if parser == 'line_json':
        import opengridview.parser.linejson
        return opengridview.parser.linejson.ParserLineJson(config)
    elif parser == 'csv':
        import opengridview.parser.csvp
        return opengridview.parser.csvp.ParserCsv(config)
    elif parser == 'autosplit':
        import opengridview.parser.autosplit
        return opengridview.parser.autosplit.ParserAutoSplit(config)
    else:
        raise Exception('cant figure out format')

  def __init__(self, config):
    self.approved_types = [int,str,float,bool,long,complex]
    self.config = config
    self.headers = None
    self.parse_headers(config.get('headers'))

  def parse_headers(self,headers):
    header_names = []
    header_types = []
    if not headers: headers=[]
    for header_item in headers:
      h_list = header_item.split(':')
      h_name = h_list[0]
      header_names.append(h_name)
      if len(h_list)>1: h_type = self.get_type(h_list[1])
      else: h_type = None
      header_types.append(h_type)
    self.header_names = header_names
    self.header_types = header_types

  def set_headers(self, titles):
    if not self.header_names:
      self.header_names = titles

  # Add a row of values as a new item
  def add_item(self,row,header_func,item_func):
    # Deduce header types if needed
    if not self.headers:
      val_types = [type(i) for i in row]
      # If we have pre-provided types, prefer it over the discovered values
      if self.header_types: val_types = self.merge_lists(self.header_types,val_types)
      # Set headers
      self.headers = zip(self.header_names,val_types)
      header_func(self.headers)
    # Convert row to acceptable types
    row = [ self.cast(z[0],z[1]) for z in zip(row,zip(*self.headers)[1]) ]
    row = self.inflate_truncate(row,len(self.headers))
    # Add item
    item_func(row)

  # Set the header types according to observed values from row. Call after setting header names
  def set_header_types(self, row_types):
    row_types = self.inflate_truncate(val_types,len(self.header_names))

  # Make l the length n, inflate using default or truncate
  def inflate_truncate(self,l,n,default=None):
    return l[:n] + [default]*(n-len(l))

  def g_type(self,value):
      if not value:
          return value
      elif type(value) in self.approved_types:
          return value
      else:
          return str(value)

  def get_type(self,type_name):
    global approved_types
    for t in self.approved_types:
      if t.__name__ == type_name:
        return t
    raise ValueError('No such type: %r' % type_name)

  def cast(self,inp,required_type):
    if inp == None:
      return None
    elif required_type == bool:
      return str(inp).lower() in ("yes", "true", "t", "1")
    else:
      return required_type(inp)

  # Merge lists like dictionary merge in ruby: Prefer the item in the first list, and if it's None defer to the same item in the second list
  def merge_lists(self,a,b):
    # Make B exactly long as A
    b = self.inflate_truncate(b,len(a))
    zipper = zip(a,b)
    return [ x[0] if x[0] else x[1] for x in zipper ]

  def read_stream(self, stream, item_func, header_func):
    raise Exception('Should be implemented in subclasses') # TODO throw proper exception



from opengridview.parser import Parser
import csv

class ParserCsv(Parser):

  def __init__(self,config):
    Parser.__init__(self,config) # Base
    self.headers = None

  def read_stream(self, stream, item_func, header_func):
    while True:
      line = stream.readline()
      if line:
        if self.config.has_key('separator'):
          reader = csv.reader([line],delimiter=self.config['separator'])
        else:
          reader = csv.reader([line])
        row = reader.next()
        if not self.header_names:
          # This is the first row. Use it to set header names
          self.set_headers(row)
          continue # Don't process as input
        else:
          self.add_item(row,header_func,item_func)
      else:
        break



from opengridview.parser import Parser
import json

class ParserLineJson(Parser):
  def __init__(self,config):
    Parser.__init__(self,config) # Base

  def read_stream(self, stream, item_func,header_func):
    while True:
      line = stream.readline()
      if line:
        item = json.loads(line)
        self.set_headers(item.keys()) # Always safe
        self.add_item([item.get(k) for k in self.header_names],header_func,item_func)
      else:
        break



import threading
from gi.repository import Gtk,Gio,Gdk
from opengridview.parser import Parser
from gi.repository import GObject
import csv
import StringIO

class Window(Gtk.ApplicationWindow):

    def __init__(self,input_stream,input_config):
        if input_config.get('title',None):
          title = 'Open-GridView - %s' % input_stream.name
        else:
          title = 'Open-GridView'
        Gtk.Window.__init__(self, title=title)

        # Set variables
        self.input_stream = input_stream
        self.input_config = input_config
        self.should_stop = False
        self.stream_end = False
        self.tree_source = None
        self.done_headers = False

        box = Gtk.Box(orientation='vertical', spacing=5)
        self.add(box)

        self.txt_filter = Gtk.Entry()
        box.pack_start(self.txt_filter, False, True, 0)

        sw = Gtk.ScrolledWindow()
        box.pack_start(sw, True, True, 0)

        self.tree_view = Gtk.TreeView()
        self.tree_view.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        sw.add(self.tree_view)

        action_group = Gtk.ActionGroup('my_actions')
        action_group.add_actions([
                ("EditMenu", None, "Edit"),
                ("EditCopy",None,'copy','<Ctrl>c',None,self.copy_clipboard_noheaders),
                ("EditCopyHeaders",None,'Copy (with headers)','<Ctrl><Shift>c',None,self.copy_clipboard_headers)
            ])
        uimanager = Gtk.UIManager()
        uimanager.add_ui_from_string("""
            <ui>
               <menubar name='MenuBar'>
                   <menu action='EditMenu'>
                       <menuitem action='EditCopy' />
                       <menuitem action='EditCopyHeaders' />
                   </menu>
               </menubar>
           </ui>
        """)
        uimanager.insert_action_group(action_group)
        accelgroup = uimanager.get_accel_group()
        menubar = uimanager.get_widget("/MenuBar")
        box.pack_start(menubar, False, False, 0)
        self.add_accel_group(accelgroup)

        self.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        # Set my position
        self.connect_after('show', lambda window:
                GObject.idle_add(self.my_set_position)
                )

    def my_set_position(self):
        screen = self.get_screen()
        monitor = screen.get_monitor_at_window(self.get_window())
        geometry = screen.get_monitor_geometry(monitor)
        self.resize( geometry.width * 0.75 , geometry.height * 0.75 )

    def copy_clipboard_noheaders(self,action):
        self._copy_clipboard(False)
    def copy_clipboard_headers(self,action):
        self._copy_clipboard(True)

    # TODO Wanted to use more sophisticated data type but cant http://stackoverflow.com/q/25151437
    def _copy_clipboard(self,include_headers):
      if not self.done_headers:
          return
      res = StringIO.StringIO()
      source,paths = self.tree_view.get_selection().get_selected_rows()
      writer = csv.writer(res)
      if include_headers:
          writer.writerow(self.headers)
      for path in paths:
        iter = source.get_iter(path)
        cells = [ source.get_value(iter,i) for i in range(0,self.tree_source.get_n_columns()) ]
        writer.writerow(cells)
      self.clipboard.set_text(res.getvalue(),-1)

    # Set headers for the table
    def set_headers(self,headers):
      header_titles,header_types = zip(*headers)
      if self.done_headers: return
      self.headers = header_titles
      self.header_types = header_types
      for idx, title in enumerate(header_titles):
        if not title: continue # Skip columns with empty titles
        r = Gtk.CellRendererText()
        c = Gtk.TreeViewColumn(title,r, text=idx)
        c.set_resizable(True)
        c.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        c.set_sort_column_id(idx)
        self.tree_view.append_column(c)
      # Create source
      header_types = map(lambda x: str if x == unicode else x, header_types)
      self.tree_source = Gtk.ListStore(*header_types)
      tree_filter = self.tree_source.filter_new()
      tree_sort = Gtk.TreeModelSort(model=tree_filter)
      self.tree_view.set_model(tree_sort)
      self.txt_filter.connect("changed", lambda x: tree_filter.refilter())
      tree_filter.set_visible_func(self.filter)
      if self.input_config.has_key('filter'):
        self.txt_filter.set_text(self.input_config['filter'])
      self.done_headers = True

    # Add new item
    def add_item(self,item):
      # assuming item is array of values
      z = zip(item,self.headers)
      if not any(((cell[0] != None) and cell[1] for cell in z)): return # Skip items that have no values in columns with names
      self.tree_source.append(item)
      if not self.get_visible():
          self.show_all()

    def filter(self,obj,iteration,b):
      filter_text = self.txt_filter.get_text()
      # simple filter
      for i in range(0,len(self.headers)):
        if filter_text in str(obj.get_value(iteration,i)): return True
      return False

    def start_read(self):
      th = threading.Thread(target=self.read_stream)
      th.daemon=True
      th.start()

    def read_stream(self):
      parser = Parser.find_parser(self.input_config)
      parser.read_stream(
              self.input_stream,
              lambda item: GObject.idle_add(self.add_item, item),
              lambda headers: GObject.idle_add(self.set_headers,headers)
              )
